import type { Article, Tag } from "./types"

// Base URL for the Drupal JSON:API
const API_BASE_URL = process.env.NEXT_PUBLIC_DRUPAL_API_URL || "http://jsonapi.local/jsonapi"

/**
 * Fetch articles from Drupal JSON:API with optional tag filtering
 */
export async function fetchArticles(tagId?: string | null): Promise<Article[]> {
  let url = `${API_BASE_URL}/node/article?include=field_image,field_tags&sort=-created`

  // Add filter for tags if specified
  if (tagId) {
    url += `&filter[field_tags.id]=${tagId}`
  }

  const response = await fetch(url, {
    headers: {
      Accept: "application/vnd.api+json",
    },
    next: { revalidate: 3600 }, // Revalidate every minute
    cache: 'no-store',
  })

  if (!response.ok) {
    throw new Error(`Failed to fetch articles: ${response.status}`)
  }

  const data = await response.json()

  // Process the JSON:API response to extract articles with included relationships
  return processArticlesResponse(data)
}

export async function fetchMainMenu() {
  let url = `${API_BASE_URL}/menu_items/main`;
  const response = await fetch(url);
  const data = await response.json();

  const menu = data.data.map((item) => ({
    id: item.id,
    title: item.attributes.title,
    url: item.attributes.url,
    description: item.attributes.description,
  }));

  if (!response.ok) {
    throw new Error(`Failed to fetch articles: ${response.status}`)
  }

  return {
    props: { menu },
    revalidate: 60, // Optional: ISR every 60 seconds
  };
}

/**
 * Fetch taxonomy terms (tags) from Drupal JSON:API
 */
export async function fetchTags(): Promise<Tag[]> {
  const url = `${API_BASE_URL}/taxonomy_term/tags?sort=name&fields[taxonomy_term--tags]=name,drupal_internal__tid`

  const response = await fetch(url, {
    headers: {
      Accept: "application/vnd.api+json",
    },
    next: { revalidate: 3600 }, // Revalidate every hour
    cache: 'no-store',
  })

  if (!response.ok) {
    throw new Error(`Failed to fetch tags: ${response.status}`)
  }

  const data = await response.json()

  // Process the JSON:API response to extract tags
  return data.data.map((item: any) => ({
    id: item.id,
    name: item.attributes.name,
    // You might want to fetch the count separately or include it in the API response
    count: null,
  }))
}

/**
 * Process the JSON:API response to extract articles with included relationships
 */
function processArticlesResponse(data: any): Article[] {
  const included = data.included || []

  // Create a map of included items for easy lookup
  const includedMap = included.reduce((acc: any, item: any) => {
    acc[`${item.type}:${item.id}`] = item
    return acc
  }, {})

  // Helper function to strip HTML tags and truncate text.
  const getCleanSummary = (item: any, maxLength = 200): string => {
    const rawSummary = item.attributes.body?.summary || item.attributes.body?.value || "";
    const plainText = rawSummary.replace(/<[^>]*>/g, '');
    return plainText.substring(0, maxLength);
  };

  return data.data.map((item: any) => {
    const article: Article = {
      id: item.id,
      title: item.attributes.title,
      summary: getCleanSummary(item),
      body: item.attributes.body?.value || "",
      created: item.attributes.created,
      image: null,
      tags: [],
    }

    // Process image relationship
    if (item.relationships.field_image?.data) {
      const imageData = item.relationships.field_image.data
      const imageIncluded = includedMap[`${imageData.type}:${imageData.id}`]

      if (imageIncluded) {
        const imageUrl = imageIncluded.attributes.uri?.url
        if (imageUrl) {
          // Handle relative or absolute URLs
          article.image = imageUrl.startsWith("http")
            ? imageUrl
            : `${process.env.NEXT_PUBLIC_DRUPAL_BASE_URL || ""}${imageUrl}`
        }
      }
    }

    // Process tags relationships
    if (item.relationships.field_tags?.data) {
      article.tags = item.relationships.field_tags.data.map((tagRef: any) => {
        const tagIncluded = includedMap[`${tagRef.type}:${tagRef.id}`]

        return {
          id: tagRef.id,
          name: tagIncluded ? tagIncluded.attributes.name : "Unknown Tag",
        }
      })
    }

    return article
  })
}

export async function fetchMenuItems() {
  const response = await fetch(`${process.env.NEXT_PUBLIC_DRUPAL_BASE_URL}/jsonapi/menu_items/main`, {
    headers: {
      'Accept': 'application/vnd.api+json',
    },
  })

  if (!response.ok) {
    throw new Error(`Failed to fetch menu items: ${response.status}`)
  }

  const data = await response.json()
  return data.data || []
}

